@@ -18,12 +18,18 @@
 # pylint: enable=line-too-long
 # pylint: disable=g-import-not-at-top,g-bad-import-order,wrong-import-position
 
-__version__ = "1.0.2"
+__version__ = "1.4.1"
 
 import pyarrow
 from pyarrow import total_allocated_bytes
 
-from . import datasets
+
+if tuple(int(i) for i in pyarrow.__version__.split(".")) < (0, 17, 1):
+    raise ImportWarning(
+        "To use `datasets`, the module `pyarrow>=0.17.1` is required, and the current version of `pyarrow` doesn't match this condition.\n"
+        "If you are running this in a Google Colab, you should probably just restart the runtime to use the right version of `pyarrow`."
+    )
+
 from .arrow_dataset import Dataset, concatenate_datasets
 from .arrow_reader import ArrowReader, ReadInstruction
 from .arrow_writer import ArrowWriter
@@ -41,8 +47,16 @@ from .features import (
     TranslationVariableLanguages,
     Value,
 )
+from .fingerprint import is_caching_enabled, set_caching_enabled
 from .info import DatasetInfo, MetricInfo
-from .inspect import inspect_dataset, inspect_metric, list_datasets, list_metrics
+from .inspect import (
+    get_dataset_config_names,
+    get_dataset_infos,
+    inspect_dataset,
+    inspect_metric,
+    list_datasets,
+    list_metrics,
+)
 from .load import import_main_class, load_dataset, load_from_disk, load_metric, prepare_module
 from .metric import Metric
 from .splits import NamedSplit, Split, SplitBase, SplitDict, SplitGenerator, SplitInfo, SubSplitInfo, percent
@@ -50,10 +64,4 @@ from .utils import *
 from .utils.tqdm_utils import disable_progress_bar
 
 
-if int(pyarrow.__version__.split(".")[1]) < 16 and int(pyarrow.__version__.split(".")[0]) == 0:
-    raise ImportWarning(
-        "To use `datasets`, the module `pyarrow>=0.16.0` is required, and the current version of `pyarrow` doesn't match this condition.\n"
-        "If you are running this in a Google Colab, you should probably just restart the runtime to use the right version of `pyarrow`."
-    )
-
-SCRIPTS_VERIONS = __version__
+SCRIPTS_VERSION = __version__
