@@ -23,17 +23,15 @@ from typing import Any, Dict, List, Optional
 import pyarrow as pa
 from tqdm.auto import tqdm
 
+from . import config
 from .features import Features, _ArrayXDExtensionType
 from .info import DatasetInfo
-from .utils.file_utils import HF_DATASETS_CACHE, hash_url_to_filename
+from .utils.file_utils import hash_url_to_filename
 from .utils.logging import WARNING, get_logger
 
 
 logger = get_logger(__name__)
 
-# Batch size constants. For more info, see:
-# https://github.com/apache/arrow/blob/master/docs/source/cpp/arrays.rst#size-limitations-and-recommendations)
-DEFAULT_MAX_BATCH_SIZE = 10_000  # hopefully it doesn't write too much at once (max is 2GB)
 type_ = type  # keep python's type function
 
 
@@ -94,9 +92,14 @@ class TypedSequence:
             type = self.type
         try:
             if isinstance(type, _ArrayXDExtensionType):
-                return pa.ExtensionArray.from_storage(type, pa.array(self.data, type.storage_dtype))
+                out = pa.ExtensionArray.from_storage(type, pa.array(self.data, type.storage_dtype))
             else:
-                return pa.array(self.data, type=type)
+                out = pa.array(self.data, type=type)
+            if trying_type and out[0].as_py() != self.data[0]:
+                raise TypeError(
+                    "Specified try_type alters data. Please check that the type/feature that you provided match the type/features of the data."
+                )
+            return out
         except (TypeError, pa.lib.ArrowInvalid) as e:  # handle type errors and overflows
             if trying_type:
                 try:
@@ -159,14 +162,14 @@ class ArrowWriter(object):
 
         self.fingerprint = fingerprint
         self.disable_nullable = disable_nullable
-        self.writer_batch_size = writer_batch_size or DEFAULT_MAX_BATCH_SIZE
+        self.writer_batch_size = writer_batch_size or config.DEFAULT_MAX_BATCH_SIZE
         self.update_features = update_features
         self.with_metadata = with_metadata
         self.unit = unit
 
         self._num_examples = 0
         self._num_bytes = 0
-        self.current_rows = []
+        self.current_rows: List[Dict[str, Any]] = []
         self.pa_writer: Optional[pa.RecordBatchStreamWriter] = None
 
     def __len__(self):
@@ -271,7 +274,7 @@ class ArrowWriter(object):
             typed_sequence = TypedSequence(batch_examples[col], type=col_type, try_type=col_try_type)
             typed_sequence_examples[col] = typed_sequence
         pa_table = pa.Table.from_pydict(typed_sequence_examples)
-        self.write_table(pa_table)
+        self.write_table(pa_table, writer_batch_size)
 
     def write_table(self, pa_table: pa.Table, writer_batch_size: Optional[int] = None):
         """Write a batch of Example to file.
@@ -340,7 +343,7 @@ class BeamWriter(object):
         self._parquet_path = os.path.splitext(path)[0]  # remove extension
         self._namespace = namespace or "default"
         self._num_examples = None
-        self._cache_dir = cache_dir or HF_DATASETS_CACHE
+        self._cache_dir = cache_dir or config.HF_DATASETS_CACHE
 
     def write_from_pcollection(self, pcoll_examples):
         """Add the final steps of the beam pipeline: write to parquet files."""
