@@ -22,17 +22,18 @@ from typing import Any, Dict, List, Optional, Tuple, Union
 
 import numpy as np
 import pyarrow as pa
-from filelock import BaseFileLock, FileLock, Timeout
 
+from . import config
 from .arrow_dataset import Dataset
 from .arrow_reader import ArrowReader
 from .arrow_writer import ArrowWriter
 from .features import Features
-from .info import MetricInfo
+from .info import DatasetInfo, MetricInfo
 from .naming import camelcase_to_snakecase
-from .utils import HF_METRICS_CACHE, copyfunc, temp_seed
+from .utils import copyfunc, temp_seed
 from .utils.download_manager import DownloadManager
 from .utils.file_utils import DownloadConfig
+from .utils.filelock import BaseFileLock, FileLock, Timeout
 from .utils.logging import get_logger
 
 
@@ -177,7 +178,7 @@ class Metric(MetricInfoMixin):
         self.max_concurrent_cache_files = max_concurrent_cache_files
 
         self.keep_in_memory = keep_in_memory
-        self._data_dir_root = os.path.expanduser(cache_dir or HF_METRICS_CACHE)
+        self._data_dir_root = os.path.expanduser(cache_dir or config.HF_METRICS_CACHE)
         self.data_dir = self._build_data_dir()
         self.seed: int = seed or np.random.get_state()[1][0]
         self.timeout: Union[int, float] = timeout
@@ -244,8 +245,8 @@ class Metric(MetricInfoMixin):
                 # We raise an error
                 if self.num_process != 1:
                     raise ValueError(
-                        f"Another metric instance is already using the local cache file at {file_path}. "
-                        f"Please specify an experiment_id (currently: {self.experiment_id}) to avoid colision "
+                        f"Error in _create_cache_file: another metric instance is already using the local cache file at {file_path}. "
+                        f"Please specify an experiment_id (currently: {self.experiment_id}) to avoid collision "
                         f"between distributed metric instances."
                     )
                 if i == self.max_concurrent_cache_files - 1:
@@ -269,6 +270,11 @@ class Metric(MetricInfoMixin):
         We wait for timeout second to let all the distributed node finish their tasks (default is 100 seconds).
         """
         if self.num_process == 1:
+            if self.cache_file_name is None:
+                raise ValueError(
+                    "Metric cache file doesn't exist. Please make sure that you call `add` or `add_batch` "
+                    "at least once before calling `compute`."
+                )
             file_paths = [self.cache_file_name]
         else:
             file_paths = [
@@ -279,13 +285,16 @@ class Metric(MetricInfoMixin):
         # Let's acquire a lock on each process files to be sure they are finished writing
         filelocks = []
         for process_id, file_path in enumerate(file_paths):
-            filelock = FileLock(file_path + ".lock")
-            try:
-                filelock.acquire(timeout=self.timeout)
-            except Timeout:
-                raise ValueError(f"Cannot acquire lock on cached file {file_path} for process {process_id}.")
+            if process_id == 0:  # process 0 already has its lock file
+                filelocks.append(self.filelock)
             else:
-                filelocks.append(filelock)
+                filelock = FileLock(file_path + ".lock")
+                try:
+                    filelock.acquire(timeout=self.timeout)
+                except Timeout:
+                    raise ValueError(f"Cannot acquire lock on cached file {file_path} for process {process_id}.")
+                else:
+                    filelocks.append(filelock)
 
         return file_paths, filelocks
 
@@ -332,12 +341,13 @@ class Metric(MetricInfoMixin):
         if self.writer is not None:
             self.writer.finalize()
         self.writer = None
-        if self.filelock is not None:
+        # release the locks of the processes > 0 so that process 0 can lock them to read + delete the data
+        if self.filelock is not None and self.process_id > 0:
             self.filelock.release()
 
         if self.keep_in_memory:
             # Read the predictions and references
-            reader = ArrowReader(path=self.data_dir, info=None)
+            reader = ArrowReader(path=self.data_dir, info=DatasetInfo(features=self.features))
             self.data = Dataset.from_buffer(self.buf_writer.getvalue())
 
         elif self.process_id == 0:
@@ -346,12 +356,12 @@ class Metric(MetricInfoMixin):
 
             # Read the predictions and references
             try:
-                reader = ArrowReader(path=self.data_dir, info=None)
+                reader = ArrowReader(path="", info=DatasetInfo(features=self.features))
                 self.data = Dataset(**reader.read_files([{"filename": f} for f in file_paths]))
             except FileNotFoundError:
                 raise ValueError(
                     "Error in finalize: another metric instance is already using the local cache file. "
-                    "Please specify an experiment_id to avoid colision between distributed metric instances."
+                    "Please specify an experiment_id to avoid collision between distributed metric instances."
                 )
 
             # Store file paths and locks and we will release/delete them after the computation.
@@ -394,10 +404,16 @@ class Metric(MetricInfoMixin):
 
             if self.buf_writer is not None:
                 self.buf_writer = None
+                del self.data
+                self.data = None
             else:
-                # Release locks and delete all the cache files
-                for filelock, file_path in zip(self.filelocks, self.file_paths):
+                # Release locks and delete all the cache files. Process 0 is released last.
+                for filelock, file_path in reversed(list(zip(self.filelocks, self.file_paths))):
                     logger.info(f"Removing {file_path}")
+                    del self.data
+                    self.data = None
+                    del self.writer
+                    self.writer = None
                     os.remove(file_path)
                     filelock.release()
 
@@ -448,8 +464,8 @@ class Metric(MetricInfoMixin):
                     self.rendez_vous_lock.acquire(timeout=timeout)
                 except TimeoutError:
                     raise ValueError(
-                        f"Another metric instance is already using the local cache file at {file_path}. "
-                        f"Please specify an experiment_id (currently: {self.experiment_id}) to avoid colision "
+                        f"Error in _init_writer: another metric instance is already using the local cache file at {file_path}. "
+                        f"Please specify an experiment_id (currently: {self.experiment_id}) to avoid collision "
                         f"between distributed metric instances."
                     )
 
@@ -493,7 +509,6 @@ class Metric(MetricInfoMixin):
         self,
         download_config: Optional[DownloadConfig] = None,
         dl_manager: Optional[DownloadManager] = None,
-        **download_and_prepare_kwargs,
     ):
         """Downloads and prepares dataset for reading.
 
@@ -528,3 +543,13 @@ class Metric(MetricInfoMixin):
     def _compute(self, *, predictions=None, references=None, **kwargs) -> Dict[str, Any]:
         """ This method defines the common API for all the metrics in the library """
         raise NotImplementedError
+
+    def __del__(self):
+        if hasattr(self, "filelock") and self.filelock is not None:
+            self.filelock.release()
+        if hasattr(self, "rendez_vous_lock") and self.rendez_vous_lock is not None:
+            self.rendez_vous_lock.release()
+        if hasattr(self, "writer"):  # in case it was already deleted
+            del self.writer
+        if hasattr(self, "data"):  # in case it was already deleted
+            del self.data
