@@ -1,5 +1,7 @@
+import importlib.util
 import os
 import tempfile
+from pathlib import PurePath
 from typing import TYPE_CHECKING, Dict, List, NamedTuple, Optional, Union
 
 import numpy as np
@@ -11,22 +13,19 @@ from .utils.logging import WARNING, get_logger
 if TYPE_CHECKING:
     from .arrow_dataset import Dataset  # noqa: F401
 
+    try:
+        from elasticsearch import Elasticsearch  # noqa: F401
 
-try:
-    import elasticsearch as es
-    import elasticsearch.helpers
-    from elasticsearch import Elasticsearch
+    except ImportError:
+        pass
+    try:
+        import faiss  # noqa: F401
 
-    _has_elasticsearch = True
-except ImportError:
-    _has_elasticsearch = False
+    except ImportError:
+        pass
 
-try:
-    import faiss
-
-    _has_faiss = True
-except ImportError:
-    _has_faiss = False
+_has_elasticsearch = importlib.util.find_spec("elasticsearch") is not None
+_has_faiss = importlib.util.find_spec("faiss") is not None
 
 
 logger = get_logger(__name__)
@@ -75,12 +74,12 @@ class BaseIndex:
             total_indices.append(indices)
         return BatchedSearchResults(total_scores, total_indices)
 
-    def save(self, file: str):
+    def save(self, file: Union[str, PurePath]):
         """Serialize the index on disk"""
         raise NotImplementedError
 
     @classmethod
-    def load(cls, file: str) -> "BaseIndex":
+    def load(cls, file: Union[str, PurePath]) -> "BaseIndex":
         """Deserialize the index from disk"""
         raise NotImplementedError
 
@@ -111,6 +110,10 @@ class ElasticSearchIndex(BaseIndex):
         ), "Please specify either `es_client` or `(host, port)`, but not both."
         host = host or "localhost"
         port = port or 9200
+
+        import elasticsearch.helpers  # noqa: need this to properly load all the es features
+        from elasticsearch import Elasticsearch  # noqa: F811
+
         self.es_client = es_client if es_client is not None else Elasticsearch([{"host": host, "port": str(port)}])
         self.es_index_name = (
             es_index_name
@@ -134,7 +137,6 @@ class ElasticSearchIndex(BaseIndex):
         Add documents to the index.
         If the documents are inside a certain column, you can specify it using the `column` argument.
         """
-        # TODO: don't rebuild if it already exists
         index_name = self.es_index_name
         index_config = self.es_index_config
         self.es_client.indices.create(index=index_name, body=index_config)
@@ -152,6 +154,8 @@ class ElasticSearchIndex(BaseIndex):
                     yield {"text": example, "_id": i}
 
         # create the ES index
+        import elasticsearch as es
+
         for ok, action in es.helpers.streaming_bulk(
             client=self.es_client,
             index=index_name,
@@ -213,9 +217,10 @@ class FaissIndex(BaseIndex):
         self.string_factory = string_factory
         self.metric_type = metric_type
         self.faiss_index = custom_index
-        assert (
-            _has_faiss
-        ), "You must install Faiss to use FaissIndex. To do so you can run `pip install faiss-cpu` or `pip install faiss-gpu`"
+        if not _has_faiss:
+            raise ImportError(
+                "You must install Faiss to use FaissIndex. To do so you can run `pip install faiss-cpu` or `pip install faiss-gpu`"
+            )
 
     def add_vectors(
         self,
@@ -229,6 +234,7 @@ class FaissIndex(BaseIndex):
         Add vectors to the index.
         If the arrays are inside a certain column, you can specify it using the `column` argument.
         """
+        import faiss  # noqa: F811
 
         # Create index
         if self.faiss_index is None:
@@ -261,8 +267,8 @@ class FaissIndex(BaseIndex):
 
         # Train
         if train_size is not None:
-            logger.info("Training the index with the first {} vectors".format(train_size))
             train_vecs = vectors[:train_size] if column is None else vectors[:train_size][column]
+            logger.info("Training the index with the first {} vectors".format(len(train_vecs)))
             self.faiss_index.train(train_vecs)
         else:
             logger.info("Ignored the training step of the faiss index as `train_size` is None.")
@@ -309,27 +315,35 @@ class FaissIndex(BaseIndex):
         scores, indices = self.faiss_index.search(queries, k)
         return BatchedSearchResults(scores, indices.astype(int))
 
-    def save(self, file: str):
+    def save(self, file: Union[str, PurePath]):
         """Serialize the FaissIndex on disk"""
+        import faiss  # noqa: F811
+
         if (
             hasattr(self.faiss_index, "device")
             and self.faiss_index.device is not None
             and self.faiss_index.device > -1
+        ) or (
+            hasattr(self.faiss_index, "getDevice")
+            and self.faiss_index.getDevice() is not None
+            and self.faiss_index.getDevice() > -1
         ):
             index = faiss.index_gpu_to_cpu(self.faiss_index)
         else:
             index = self.faiss_index
-        faiss.write_index(index, file)
+        faiss.write_index(index, str(file))
 
     @classmethod
     def load(
         cls,
-        file: str,
+        file: Union[str, PurePath],
         device: Optional[int] = None,
     ) -> "FaissIndex":
         """Deserialize the FaissIndex from disk"""
+        import faiss  # noqa: F811
+
         faiss_index = cls(device=device)
-        index = faiss.read_index(file)
+        index = faiss.read_index(str(file))
         if faiss_index.device is not None and faiss_index.device > -1:
             faiss_index.faiss_res = faiss.StandardGpuResources()
             index = faiss.index_cpu_to_gpu(faiss_index.faiss_res, faiss_index.device, index)
@@ -385,21 +399,22 @@ class IndexableMixin:
         - For `string factory`: https://github.com/facebookresearch/faiss/wiki/The-index-factory
 
         Args:
-            column (`str`): The column of the vectors to add to the index.
-            index_name (Optional `str`): The index_name/identifier of the index. This is the index_name that is used to call `.get_nearest` or `.search`.
+            column (:obj:`str`): The column of the vectors to add to the index.
+            index_name (Optional :obj:`str`): The index_name/identifier of the index. This is the index_name that is used to call `.get_nearest` or `.search`.
                 By defaul it corresponds to `column`.
-            device (Optional `int`): If not None, this is the index of the GPU to use. By default it uses the CPU.
-            string_factory (Optional `str`): This is passed to the index factory of Faiss to create the index. Default index class is IndexFlatIP.
-            metric_type (Optional `int`): Type of metric. Ex: faiss.faiss.METRIC_INNER_PRODUCT or faiss.METRIC_L2.
-            custom_index (Optional `faiss.Index`): Custom Faiss index that you already have instantiated and configured for your needs.
-            train_size (Optional `int`): If the index needs a training step, specifies how many vectors will be used to train the index.
-            faiss_verbose (`bool`, defaults to False): Enable the verbosity of the Faiss index.
+            device (Optional :obj:`int`): If not None, this is the index of the GPU to use. By default it uses the CPU.
+            string_factory (Optional :obj:`str`): This is passed to the index factory of Faiss to create the index. Default index class is IndexFlatIP.
+            metric_type (Optional :obj:`int`): Type of metric. Ex: faiss.faiss.METRIC_INNER_PRODUCT or faiss.METRIC_L2.
+            custom_index (Optional :obj:`faiss.Index`): Custom Faiss index that you already have instantiated and configured for your needs.
+            train_size (Optional :obj:`int`): If the index needs a training step, specifies how many vectors will be used to train the index.
+            faiss_verbose (:obj:`bool`, defaults to False): Enable the verbosity of the Faiss index.
         """
         index_name = index_name if index_name is not None else column
-        self._indexes[index_name] = FaissIndex(
+        faiss_index = FaissIndex(
             device=device, string_factory=string_factory, metric_type=metric_type, custom_index=custom_index
         )
-        self._indexes[index_name].add_vectors(self, column=column, train_size=train_size, faiss_verbose=faiss_verbose)
+        faiss_index.add_vectors(self, column=column, train_size=train_size, faiss_verbose=faiss_verbose)
+        self._indexes[index_name] = faiss_index
 
     def add_faiss_index_from_external_arrays(
         self,
@@ -419,29 +434,28 @@ class IndexableMixin:
         - For `string factory`: https://github.com/facebookresearch/faiss/wiki/The-index-factory
 
         Args:
-            external_arrays (`np.array`): If you want to use arrays from outside the lib for the index, you can set `external_arrays`.
+            external_arrays (:obj:`np.array`): If you want to use arrays from outside the lib for the index, you can set `external_arrays`.
                 It will use `external_arrays` to create the Faiss index instead of the arrays in the given `column`.
-            index_name (`str`): The index_name/identifier of the index. This is the index_name that is used to call `.get_nearest` or `.search`.
-            device (Optional `int`): If not None, this is the index of the GPU to use. By default it uses the CPU.
-            string_factory (Optional `str`): This is passed to the index factory of Faiss to create the index. Default index class is IndexFlatIP.
-            metric_type (Optional `int`): Type of metric. Ex: faiss.faiss.METRIC_INNER_PRODUCT or faiss.METRIC_L2.
-            custom_index (Optional `faiss.Index`): Custom Faiss index that you already have instantiated and configured for your needs.
-            train_size (Optional `int`): If the index needs a training step, specifies how many vectors will be used to train the index.
-            faiss_verbose (`bool`, defaults to False): Enable the verbosity of the Faiss index.
+            index_name (:obj:`str`): The index_name/identifier of the index. This is the index_name that is used to call `.get_nearest` or `.search`.
+            device (Optional :obj:`int`): If not None, this is the index of the GPU to use. By default it uses the CPU.
+            string_factory (Optional :obj:`str`): This is passed to the index factory of Faiss to create the index. Default index class is IndexFlatIP.
+            metric_type (Optional :obj:`int`): Type of metric. Ex: faiss.faiss.METRIC_INNER_PRODUCT or faiss.METRIC_L2.
+            custom_index (Optional :obj:`faiss.Index`): Custom Faiss index that you already have instantiated and configured for your needs.
+            train_size (Optional :obj:`int`): If the index needs a training step, specifies how many vectors will be used to train the index.
+            faiss_verbose (:obj:`bool`, defaults to False): Enable the verbosity of the Faiss index.
         """
-        self._indexes[index_name] = FaissIndex(
+        faiss_index = FaissIndex(
             device=device, string_factory=string_factory, metric_type=metric_type, custom_index=custom_index
         )
-        self._indexes[index_name].add_vectors(
-            external_arrays, column=None, train_size=train_size, faiss_verbose=faiss_verbose
-        )
+        faiss_index.add_vectors(external_arrays, column=None, train_size=train_size, faiss_verbose=faiss_verbose)
+        self._indexes[index_name] = faiss_index
 
-    def save_faiss_index(self, index_name: str, file: str):
+    def save_faiss_index(self, index_name: str, file: Union[str, PurePath]):
         """Save a FaissIndex on disk
 
         Args:
-            index_name (`str`): The index_name/identifier of the index. This is the index_name that is used to call `.get_nearest` or `.search`.
-            file (`str`): The path to the serialized faiss index on disk.
+            index_name (:obj:`str`): The index_name/identifier of the index. This is the index_name that is used to call `.get_nearest` or `.search`.
+            file (:obj:`str`): The path to the serialized faiss index on disk.
         """
         index = self.get_index(index_name)
         if not isinstance(index, FaissIndex):
@@ -452,16 +466,16 @@ class IndexableMixin:
     def load_faiss_index(
         self,
         index_name: str,
-        file: str,
+        file: Union[str, PurePath],
         device: Optional[int] = None,
     ):
         """Load a FaissIndex from disk.
         If you want to do additional configurations, you can have access to the faiss index object by doing `.get_index(index_name).faiss_index` to make it fit your needs
 
         Args:
-            index_name (`str`): The index_name/identifier of the index. This is the index_name that is used to call `.get_nearest` or `.search`.
-            file (`str`): The path to the serialized faiss index on disk.
-            device (Optional `int`): If not None, this is the index of the GPU to use. By default it uses the CPU.
+            index_name (:obj:`str`): The index_name/identifier of the index. This is the index_name that is used to call `.get_nearest` or `.search`.
+            file (:obj:`str`): The path to the serialized faiss index on disk.
+            device (Optional :obj:`int`): If not None, this is the index of the GPU to use. By default it uses the CPU.
         """
         index = FaissIndex.load(file, device=device)
         assert index.faiss_index.ntotal == len(
@@ -485,35 +499,96 @@ class IndexableMixin:
         """Add a text index using ElasticSearch for fast retrieval.
 
         Args:
-            column (`str`): The column of the documents to add to the index.
-            index_name (Optional `str`): The index_name/identifier of the index. This is the index name that is used to call `.get_nearest` or `.search`.
+            column (:obj:`str`): The column of the documents to add to the index.
+            index_name (Optional :obj:`str`): The index_name/identifier of the index. This is the index name that is used to call `.get_nearest` or `.search`.
                 By defaul it corresponds to `column`.
-            documents (`Union[List[str], datasets.Dataset]`): The documents to index. It can be a `datasets.Dataset`.
-            es_client (`elasticsearch.Elasticsearch`): The elasticsearch client used to create the index.
-            es_index_name (Optional `str`): The elasticsearch index name used to create the index.
-            es_index_config (Optional `dict`): The configuration of the elasticsearch index.
-                Default config is
-                {
-                    "settings": {
-                        "number_of_shards": 1,
-                        "analysis": {"analyzer": {"stop_standard": {"type": "standard", " stopwords": "_english_"}}},
-                    },
-                    "mappings": {
-                        "properties": {
-                            "text": {"type": "text", "analyzer": "standard", "similarity": "BM25"},
-                        }
-                    },
-                }
+            host (Optional :obj:`str`, defaults to localhost):
+                host of where ElasticSearch is running
+            port (Optional :obj:`str`, defaults to 9200):
+                port of where ElasticSearch is running
+            es_client (Optional :obj:`elasticsearch.Elasticsearch`):
+                The elasticsearch client used to create the index if host and port are None.
+            es_index_name (Optional :obj:`str`): The elasticsearch index name used to create the index.
+            es_index_config (Optional :obj:`dict`):
+                The configuration of the elasticsearch index.
+                Default config is:
+
+        Config::
+
+            {
+                "settings": {
+                    "number_of_shards": 1,
+                    "analysis": {"analyzer": {"stop_standard": {"type": "standard", " stopwords": "_english_"}}},
+                },
+                "mappings": {
+                    "properties": {
+                        "text": {
+                            "type": "text",
+                            "analyzer": "standard",
+                            "similarity": "BM25"
+                        },
+                    }
+                },
+            }
         """
         index_name = index_name if index_name is not None else column
-        self._indexes[index_name] = ElasticSearchIndex(host, port, es_client, es_index_name, es_index_config)
-        self._indexes[index_name].add_documents(self, column=column)
+        es_index = ElasticSearchIndex(
+            host=host, port=port, es_client=es_client, es_index_name=es_index_name, es_index_config=es_index_config
+        )
+        es_index.add_documents(self, column=column)
+        self._indexes[index_name] = es_index
+
+    def load_elasticsearch_index(
+        self,
+        index_name: str,
+        es_index_name: str,
+        host: Optional[str] = None,
+        port: Optional[int] = None,
+        es_client: Optional["Elasticsearch"] = None,
+        es_index_config: Optional[dict] = None,
+    ):
+        """Load an existing text index using ElasticSearch for fast retrieval.
+
+        Args:
+            index_name (:obj:`str`): The index_name/identifier of the index. This is the index name that is used to call `.get_nearest` or `.search`.
+            es_index_name (`:obj:str`): The name of elasticsearch index to load.
+            host (Optional :obj:`str`, defaults to localhost):
+                host of where ElasticSearch is running
+            port (Optional `:obj:str`, defaults to 9200):
+                port of where ElasticSearch is running
+            es_client (Optional :obj:`elasticsearch.Elasticsearch`):
+                The elasticsearch client used to create the index if host and port are None.
+            es_index_config (Optional :obj:`dict`):
+                The configuration of the elasticsearch index.
+                Default config is:
+
+        Config::
+
+            {
+                "settings": {
+                    "number_of_shards": 1,
+                    "analysis": {"analyzer": {"stop_standard": {"type": "standard", " stopwords": "_english_"}}},
+                },
+                "mappings": {
+                    "properties": {
+                        "text": {
+                            "type": "text",
+                            "analyzer": "standard",
+                            "similarity": "BM25"
+                        },
+                    }
+                },
+            }
+        """
+        self._indexes[index_name] = ElasticSearchIndex(
+            host=host, port=port, es_client=es_client, es_index_name=es_index_name, es_index_config=es_index_config
+        )
 
     def drop_index(self, index_name: str):
         """Drop the index with the specified column.
 
         Args:
-            index_name (`str`): The index_name/identifier of the index.
+            index_name (:obj:`str`): The index_name/identifier of the index.
         """
         del self._indexes[index_name]
 
@@ -521,13 +596,13 @@ class IndexableMixin:
         """Find the nearest examples indices in the dataset to the query.
 
         Args:
-            index_name (`str`): The name/identifier of the index.
-            query (`Union[str, np.ndarray]`): The query as a string if `index_name` is a text index or as a numpy array if `index_name` is a vector index.
-            k (`int`): The number of examples to retrieve.
+            index_name (:obj:`str`): The name/identifier of the index.
+            query (:obj:`Union[str, np.ndarray]`): The query as a string if `index_name` is a text index or as a numpy array if `index_name` is a vector index.
+            k (:obj:`int`): The number of examples to retrieve.
 
         Ouput:
-            scores (`List[List[float]`): The retrieval scores of the retrieved examples.
-            indices (`List[List[int]]`): The indices of the retrieved examples.
+            scores (:obj:`List[List[float]`): The retrieval scores of the retrieved examples.
+            indices (:obj:`List[List[int]]`): The indices of the retrieved examples.
         """
         self._check_index_is_initialized(index_name)
         return self._indexes[index_name].search(query, k)
@@ -536,13 +611,13 @@ class IndexableMixin:
         """Find the nearest examples indices in the dataset to the query.
 
         Args:
-            index_name (`str`): The index_name/identifier of the index.
-            queries (`Union[List[str], np.ndarray]`): The queries as a list of strings if `index_name` is a text index or as a numpy array if `index_name` is a vector index.
-            k (`int`): The number of examples to retrieve per query.
+            index_name (:obj:`str`): The index_name/identifier of the index.
+            queries (:obj:`Union[List[str], np.ndarray]`): The queries as a list of strings if `index_name` is a text index or as a numpy array if `index_name` is a vector index.
+            k (`:obj:int`): The number of examples to retrieve per query.
 
         Ouput:
-            total_scores (`List[List[float]`): The retrieval scores of the retrieved examples per query.
-            total_indices (`List[List[int]]`): The indices of the retrieved examples per query.
+            total_scores (:obj:`List[List[float]`): The retrieval scores of the retrieved examples per query.
+            total_indices (:obj:`List[List[int]]`): The indices of the retrieved examples per query.
         """
         self._check_index_is_initialized(index_name)
         return self._indexes[index_name].search_batch(queries, k)
@@ -553,13 +628,13 @@ class IndexableMixin:
         """Find the nearest examples in the dataset to the query.
 
         Args:
-            index_name (`str`): The index_name/identifier of the index.
-            query (`Union[str, np.ndarray]`): The query as a string if `index_name` is a text index or as a numpy array if `index_name` is a vector index.
-            k (`int`): The number of examples to retrieve.
+            index_name (:obj:`str`): The index_name/identifier of the index.
+            query (:obj:`Union[str, np.ndarray]`): The query as a string if `index_name` is a text index or as a numpy array if `index_name` is a vector index.
+            k (:obj:`int`): The number of examples to retrieve.
 
         Ouput:
-            scores (`List[float]`): The retrieval scores of the retrieved examples.
-            examples (`dict`): The retrieved examples.
+            scores (:obj:`List[float]`): The retrieval scores of the retrieved examples.
+            examples (:obj:`dict`): The retrieved examples.
         """
         self._check_index_is_initialized(index_name)
         scores, indices = self.search(index_name, query, k)
@@ -571,9 +646,9 @@ class IndexableMixin:
         """Find the nearest examples in the dataset to the query.
 
         Args:
-            index_name (`str`): The index_name/identifier of the index.
-            queries (`Union[List[str], np.ndarray]`): The queries as a list of strings if `index_name` is a text index or as a numpy array if `index_name` is a vector index.
-            k (`int`): The number of examples to retrieve per query.
+            index_name (:obj:`str`): The index_name/identifier of the index.
+            queries (:obj:`Union[List[str], np.ndarray]`): The queries as a list of strings if `index_name` is a text index or as a numpy array if `index_name` is a vector index.
+            k (:obj:`int`): The number of examples to retrieve per query.
 
         Ouput:
             total_scores (`List[List[float]`): The retrieval scores of the retrieved examples per query.
