@@ -20,10 +20,12 @@ from __future__ import absolute_import, division, print_function
 
 import abc
 import collections
+import re
 from dataclasses import dataclass, field
 from typing import Dict, List, Optional, Union
 
 from .arrow_reader import FileInstructions, make_file_instructions
+from .naming import _split_re
 from .utils.py_utils import NonMutableDict
 
 
@@ -32,7 +34,7 @@ class SplitInfo:
     name: str = ""
     num_bytes: int = 0
     num_examples: int = 0
-    dataset_name: str = None
+    dataset_name: Optional[str] = None
 
     @property
     def file_instructions(self):
@@ -74,7 +76,7 @@ class SplitBase(metaclass=abc.ABCMeta):
     """Abstract base class for Split compositionality.
 
     See the
-    [guide on splits](https://github.com/huggingface/datasets/tree/master/docs/splits.md)
+    [guide on splits](https://github.com/huggingface/datasets/blob/master/docs/source/splits.rst)
     for more information.
 
     There are three parts to the composition:
@@ -252,7 +254,7 @@ class PercentSlice(metaclass=PercentSliceMeta):
     """Syntactic sugar for defining slice subsplits: `datasets.percent[75:-5]`.
 
     See the
-    [guide on splits](https://github.com/huggingface/datasets/tree/master/docs/splits.md)
+    [guide on splits](https://github.com/huggingface/datasets/blob/master/docs/source/splits.rst)
     for more information.
     """
     # pylint: enable=line-too-long
@@ -347,6 +349,10 @@ class NamedSplit(SplitBase):
 
     def __init__(self, name):
         self._name = name
+        split_names_from_instruction = [split_instruction.split("[")[0] for split_instruction in name.split("+")]
+        for split_name in split_names_from_instruction:
+            if not re.match(_split_re, split_name):
+                raise ValueError(f"Split name should match '{_split_re}'' but got '{split_name}'.")
 
     def __str__(self):
         return self._name
@@ -404,7 +410,7 @@ class Split(object):
     Note: All splits, including compositions inherit from `datasets.SplitBase`
 
     See the
-    [guide on splits](https://github.com/huggingface/datasets/tree/master/docs/splits.md)
+    [guide on splits](https://github.com/huggingface/datasets/blob/master/docs/source/splits.rst)
     for more information.
     """
     # pylint: enable=line-too-long
@@ -528,7 +534,7 @@ class SplitDict(dict):
             split_infos = list(split_infos.values())
 
         if dataset_name is None:
-            dataset_name = split_infos[0]["dataset_name"]
+            dataset_name = split_infos[0]["dataset_name"] if split_infos else None
 
         split_dict = cls(dataset_name=dataset_name)
 
@@ -570,4 +576,5 @@ class SplitGenerator:
 
     def __post_init__(self):
         self.name = str(self.name)  # Make sure we convert NamedSplits in strings
+        NamedSplit(self.name)  # check that it's a valid split name
         self.split_info = SplitInfo(name=self.name)
