@@ -22,12 +22,14 @@ import contextlib
 import functools
 import itertools
 import os
+import pickle
+import sys
 import types
 from io import BytesIO as StringIO
 from multiprocessing import Pool, RLock
 from shutil import disk_usage
-from types import CodeType
-from typing import Optional
+from types import CodeType, FunctionType
+from typing import Callable, ClassVar, Generic, Optional, Tuple, Union
 
 import dill
 import numpy as np
@@ -36,6 +38,16 @@ from tqdm import tqdm
 from .logging import INFO, WARNING, get_logger, get_verbosity, set_verbosity_warning
 
 
+try:  # pragma: no branch
+    import typing_extensions as _typing_extensions
+    from typing_extensions import Final, Literal
+except ImportError:
+    _typing_extensions = Literal = Final = None
+
+
+logger = get_logger(__name__)
+
+
 # NOTE: When used on an instance method, the cache is shared across all
 # instances and IS NOT per-instance.
 # See
@@ -43,8 +55,6 @@ from .logging import INFO, WARNING, get_logger, get_verbosity, set_verbosity_war
 # For @property methods, use @memoized_property below.
 memoize = functools.lru_cache
 
-logger = get_logger(__name__)
-
 
 def size_str(size_in_bytes):
     """Returns a human readable size string.
@@ -70,20 +80,6 @@ def size_str(size_in_bytes):
     return "{} {}".format(int(size_in_bytes), "bytes")
 
 
-def is_notebook():
-    """Returns True if running in a notebook (Colab, Jupyter) environement."""
-    # Inspired from the tfdm autonotebook code
-    try:
-        from IPython import get_ipython  # pylint: disable=import-outside-toplevel,g-import-not-at-top
-
-        if "IPKernelApp" not in get_ipython().config:
-            return False  # Run in a IPython terminal
-    except:  # noqa: E722
-        return False
-    else:
-        return True
-
-
 @contextlib.contextmanager
 def temporary_assignment(obj, attr, value):
     """Temporarily assign obj.attr to value."""
@@ -137,23 +133,6 @@ class classproperty(property):  # pylint: disable=invalid-name
         return self.fget.__get__(None, objtype)()
 
 
-class memoized_property(property):  # pylint: disable=invalid-name
-    """Descriptor that mimics @property but caches output in member variable."""
-
-    def __get__(self, obj, objtype=None):
-        # See https://docs.python.org/3/howto/descriptor.html#properties
-        if obj is None:
-            return self
-        if self.fget is None:
-            raise AttributeError("unreadable attribute")
-        attr = "__cached_" + self.fget.__name__
-        cached = getattr(obj, attr, None)
-        if cached is None:
-            cached = self.fget(obj)
-            setattr(obj, attr, cached)
-        return cached
-
-
 def _single_map_nested(args):
     """Apply a function recursively to each element of a nested data struct."""
     function, data_struct, types, rank, disable_tqdm = args
@@ -287,37 +266,18 @@ def flatten_nest_dict(d):
     return flat_dict
 
 
-def flatten_nested(data_struct):
-    """Flatten data struct of obj or `list`/`dict` of obj"""
-    if isinstance(data_struct, dict):
-        data_struct = list(flatten_nest_dict(data_struct).values())
-        if data_struct and isinstance(data_struct[0], (list, tuple)):
-            data_struct = [x for sublist in data_struct for x in sublist]
-    if isinstance(data_struct, (list, tuple)):
-        return data_struct
-    # Singleton
-    return [data_struct]
+class NestedDataStructure:
+    def __init__(self, data=None):
+        self.data = data if data is not None else []
 
-
-def datasets_dir():
-    """Path to datasets directory."""
-    return os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
-
-
-class abstractclassmethod(classmethod):  # pylint: disable=invalid-name
-    """Decorate a method to mark it as an abstract @classmethod."""
-
-    __isabstractmethod__ = True
-
-    def __init__(self, fn):
-        fn.__isabstractmethod__ = True
-        super(abstractclassmethod, self).__init__(fn)
-
-
-def get_datasets_path(relative_path):
-    """Returns absolute path to file given path relative to datasets root."""
-    path = os.path.join(datasets_dir(), relative_path)
-    return path
+    def flatten(self, data=None):
+        data = data if data is not None else self.data
+        if isinstance(data, dict):
+            return self.flatten(list(data.values()))
+        elif isinstance(data, (list, tuple)):
+            return [flattened for item in data for flattened in self.flatten(item)]
+        else:
+            return [data]
 
 
 def has_sufficient_disk_space(needed_bytes, directory="."):
@@ -333,6 +293,19 @@ class Pickler(dill.Pickler):
 
     dispatch = dill._dill.MetaCatchingDict(dill.Pickler.dispatch.copy())
 
+    def save_global(self, obj, name=None):
+        if sys.version_info[:2] < (3, 7) and _CloudPickleTypeHintFix._is_parametrized_type_hint(
+            obj
+        ):  # noqa  # pragma: no branch
+            # Parametrized typing constructs in Python < 3.7 are not compatible
+            # with type checks and ``isinstance`` semantics. For this reason,
+            # it is easier to detect them using a duck-typing-based check
+            # (``_is_parametrized_type_hint``) than to populate the Pickler's
+            # dispatch with type-specific savers.
+            _CloudPickleTypeHintFix._save_parametrized_type_hint(self, obj)
+        else:
+            dill.Pickler.save_global(self, obj, name=name)
+
 
 def dump(obj, file):
     """pickle an object to a file"""
@@ -376,6 +349,47 @@ def pklregister(t):
     return proxy
 
 
+class _CloudPickleTypeHintFix:
+    """
+    Type hints can't be properly pickled in python < 3.7
+    CloudPickle provided a way to make it work in older versions.
+    This class provide utilities to fix pickling of type hints in older versions.
+    from https://github.com/cloudpipe/cloudpickle/pull/318/files
+    """
+
+    def _is_parametrized_type_hint(obj):
+        # This is very cheap but might generate false positives.
+        origin = getattr(obj, "__origin__", None)  # typing Constructs
+        values = getattr(obj, "__values__", None)  # typing_extensions.Literal
+        type_ = getattr(obj, "__type__", None)  # typing_extensions.Final
+        return origin is not None or values is not None or type_ is not None
+
+    def _create_parametrized_type_hint(origin, args):
+        return origin[args]
+
+    def _save_parametrized_type_hint(pickler, obj):
+        # The distorted type check sematic for typing construct becomes:
+        # ``type(obj) is type(TypeHint)``, which means "obj is a
+        # parametrized TypeHint"
+        if type(obj) is type(Literal):  # pragma: no branch
+            initargs = (Literal, obj.__values__)
+        elif type(obj) is type(Final):  # pragma: no branch
+            initargs = (Final, obj.__type__)
+        elif type(obj) is type(ClassVar):
+            initargs = (ClassVar, obj.__type__)
+        elif type(obj) in [type(Union), type(Tuple), type(Generic)]:
+            initargs = (obj.__origin__, obj.__args__)
+        elif type(obj) is type(Callable):
+            args = obj.__args__
+            if args[0] is Ellipsis:
+                initargs = (obj.__origin__, args)
+            else:
+                initargs = (obj.__origin__, (list(args[:-1]), args[-1]))
+        else:  # pragma: no cover
+            raise pickle.PicklingError("Datasets pickle Error: Unknown type {}".format(type(obj)))
+        pickler.save_reduce(_CloudPickleTypeHintFix._create_parametrized_type_hint, initargs, obj=obj)
+
+
 @pklregister(CodeType)
 def _save_code(pickler, obj):
     """
@@ -384,13 +398,18 @@ def _save_code(pickler, obj):
     of functions created in notebooks or shells for example.
     """
     dill._dill.log.info("Co: %s" % obj)
+    # The filename of a function is the .py file where it is defined.
     # Filenames of functions created in notebooks or shells start with '<'
     # ex: <ipython-input-13-9ed2afe61d25> for ipython, and <stdin> for shell
     # Moreover lambda functions have a special name: '<lambda>'
     # ex: (lambda x: x).__code__.co_name == "<lambda>"  # True
+    # For the hashing mechanism we ignore where the function has been defined
+    # More specifically:
+    # - we ignore the filename of special functions (filename starts with '<')
+    # - we always ignore the line number
     # Only those two lines are different from the original implementation:
     co_filename = "" if obj.co_filename.startswith("<") or obj.co_name == "<lambda>" else obj.co_filename
-    co_firstlineno = 1 if obj.co_filename.startswith("<") or obj.co_name == "<lambda>" else obj.co_firstlineno
+    co_firstlineno = 1
     # The rest is the same as in the original dill implementation
     if dill._dill.PY3:
         if hasattr(obj, "co_posonlyargcount"):
@@ -452,6 +471,80 @@ def _save_code(pickler, obj):
     return
 
 
+@pklregister(FunctionType)
+def save_function(pickler, obj):
+    """
+    From dill._dill.save_function
+    This is a modified version that make globs deterministic since the order of
+    the keys in the output dictionary of globalvars can change.
+    """
+    if not dill._dill._locate_function(obj):
+        dill._dill.log.info("F1: %s" % obj)
+        if getattr(pickler, "_recurse", False):
+            # recurse to get all globals referred to by obj
+            globalvars = dill.detect.globalvars
+            globs = globalvars(obj, recurse=True, builtin=True)
+            if id(obj) in dill._dill.stack:
+                globs = obj.__globals__ if dill._dill.PY3 else obj.func_globals
+        else:
+            globs = obj.__globals__ if dill._dill.PY3 else obj.func_globals
+        # globs is a dictionary with keys = var names (str) and values = python objects
+        # however the dictionary is not always loaded in the same order
+        # therefore we have to sort the keys to make deterministic.
+        # This is important to make `dump` deterministic.
+        # Only this line is different from the original implementation:
+        globs = {k: globs[k] for k in sorted(globs.keys())}
+        # The rest is the same as in the original dill implementation
+        _byref = getattr(pickler, "_byref", None)
+        _recurse = getattr(pickler, "_recurse", None)
+        _memo = (id(obj) in dill._dill.stack) and (_recurse is not None)
+        dill._dill.stack[id(obj)] = len(dill._dill.stack), obj
+        if dill._dill.PY3:
+            _super = ("super" in getattr(obj.__code__, "co_names", ())) and (_byref is not None)
+            if _super:
+                pickler._byref = True
+            if _memo:
+                pickler._recurse = False
+            fkwdefaults = getattr(obj, "__kwdefaults__", None)
+            pickler.save_reduce(
+                dill._dill._create_function,
+                (obj.__code__, globs, obj.__name__, obj.__defaults__, obj.__closure__, obj.__dict__, fkwdefaults),
+                obj=obj,
+            )
+        else:
+            _super = (
+                ("super" in getattr(obj.func_code, "co_names", ()))
+                and (_byref is not None)
+                and getattr(pickler, "_recurse", False)
+            )
+            if _super:
+                pickler._byref = True
+            if _memo:
+                pickler._recurse = False
+            pickler.save_reduce(
+                dill._dill._create_function,
+                (obj.func_code, globs, obj.func_name, obj.func_defaults, obj.func_closure, obj.__dict__),
+                obj=obj,
+            )
+        if _super:
+            pickler._byref = _byref
+        if _memo:
+            pickler._recurse = _recurse
+        if (
+            dill._dill.OLDER
+            and not _byref
+            and (_super or (not _super and _memo) or (not _super and not _memo and _recurse))
+        ):
+            pickler.clear_memo()
+        dill._dill.log.info("# F1")
+    else:
+        dill._dill.log.info("F2: %s" % obj)
+        name = getattr(obj, "__qualname__", getattr(obj, "__name__", None))
+        dill._dill.StockPickler.save_global(pickler, obj, name=name)
+        dill._dill.log.info("# F2")
+    return
+
+
 def copyfunc(func):
     return types.FunctionType(func.__code__, func.__globals__, func.__name__, func.__defaults__, func.__closure__)
 
